from   collections.abc   import Iterator
from   itertools         import groupby
import json
from   operator          import itemgetter
from   textwrap          import indent
import click
from   packaging.version import parse

def obj_option(*args, **kwargs):
    """
    Like `click.option`, but sets an attribute on ``ctx.obj`` instead of
    creating a parameter
    """
    def callback(ctx, param, value):
        setattr(ctx.obj, param.name, value)
    return click.option(*args, callback=callback, expose_value=False, **kwargs)

pre_opt = obj_option(
    '--pre/--no-pre',
    default=False,
    help='Show prerelease versions',
    show_default=True,
)

all_opt = obj_option(
    '-A',
    '--all-versions/--latest-version',
    default=False,
    help='Show all versions/only the latest version when no version is specified [default: latest]',
)

sort_opt = obj_option(
    '--newest/--highest',
    default=False,
    help='Does "latest" mean "newest" or "highest"? [default: highest]',
)

def package_args(versioned=True):
    if versioned:
        def wrapper(f):
            return all_opt(sort_opt(pre_opt(click.argument(
                'packages',
                nargs=-1,
                callback=lambda ctx, param, value:
                            ctx.obj.lookup_package_version(value),
            )(f))))
        return wrapper
    else:
        return click.argument(
            'packages',
            nargs=-1,
            callback=lambda ctx, param, value: ctx.obj.lookup_package(value),
        )

def dumps(obj):
    if isinstance(obj, Iterator):
        obj = list(obj)
    return json.dumps(obj, sort_keys=True, indent=4, ensure_ascii=False)

def clean_pypi_dict(d):
    return {
        k: (None if v in ('', 'UNKNOWN') else v)
        for k,v in d.items() if not k.startswith(('cheesecake', '_pypi'))
    }

def squish_versions(releases):
    """
    Given a list of `dict`s containing (at least) ``"name"`` and ``"version"``
    fields, return for each name the `dict` with the highest version.

    It is assumed that `dict`s with the same name are always adjacent.
    """
    for _, versions in groupby(releases, itemgetter("name")):
        yield max(versions, key=lambda v: parse(v["version"]))

class JSONLister:
    def __init__(self):
        self.first = True

    def __enter__(self):
        click.echo('[', nl=False)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if not self.first:
            click.echo()
        click.echo(']')
        return False

    def append(self, obj):
        if self.first:
            click.echo()
            self.first = False
        else:
            click.echo(',')
        click.echo(indent(dumps(obj), ' '*4), nl=False)


class JSONMapper:
    def __init__(self):
        self.first = True

    def __enter__(self):
        click.echo('{', nl=False)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if not self.first:
            click.echo()
        click.echo('}')
        return False

    def append(self, key, value):
        if self.first:
            click.echo()
            self.first = False
        else:
            click.echo(',')
        click.echo(indent(json.dumps(key)+': '+dumps(value), ' '*4), nl=False)



import platform
from   xmlrpc.client     import ServerProxy
import click
from   packaging.version import parse
import requests
from   .                 import __url__, __version__

USER_AGENT = 'qypi/{} ({}) requests/{} {}/{}'.format(
    __version__,
    __url__,
    requests.__version__,
    platform.python_implementation(),
    platform.python_version(),
)

class QyPI:
    def __init__(self, index_url):
        self.index_url = index_url
        self.s = None
        self.xsp = None
        self.pre = False
        self.newest = False
        self.all_versions = False
        self.errmsgs = []

    def get(self, *path):
        if self.s is None:
            self.s = requests.Session()
            self.s.headers["User-Agent"] = USER_AGENT
        return self.s.get(self.index_url.rstrip('/') + '/' + '/'.join(path))

    def get_package(self, package):
        r = self.get(package, 'json')
        # Unlike the XML-RPC API, the JSON API accepts package names regardless
        # of normalization
        if r.status_code == 404:
            raise QyPIError(package + ': package not found')
        r.raise_for_status()
        return r.json()

    def get_latest_version(self, package):
        pkg = self.get_package(package)
        releases = {
            (parse(rel), rel): first_upload(files)
            # The unparsed version string needs to be kept around because the
            # alternative approach (stringifying the Version object once
            # comparisons are done) can result in a different string (e.g.,
            # "2001.01.01" becomes "2001.1.1"), leading to a 404.
            for rel, files in pkg["releases"].items()
        }
        candidates = releases.keys()
        if not self.pre and any(not v[0].is_prerelease for v in candidates):
            candidates = filter(lambda v: not v[0].is_prerelease, candidates)
        if self.newest:
            latest = max(
                filter(releases.__getitem__, candidates),
                key=releases.__getitem__,
                default=None,
            )
        else:
            latest = max(candidates, default=None)
        if latest is None:
            raise QyPIError(package + ': no suitable versions available')
        latest = latest[1]
        if pkg["info"]["version"] == latest:
            return pkg
        else:
            return self.get_version(package, latest)

    def get_version(self, package, version):
        r = self.get(package, version, 'json')
        if r.status_code == 404:
            raise QyPIError('{}: version {} not found'.format(package, version))
        r.raise_for_status()
        return r.json()

    def xmlrpc(self, method, *args, **kwargs):
        if self.xsp is None:
            self.xsp = ServerProxy(self.index_url)
        return getattr(self.xsp, method)(*args, **kwargs)

    def lookup_package(self, args):
        for name in args:
            try:
                yield self.get_package(name)
            except QyPIError as e:
                self.errmsgs.append(str(e))

    def lookup_package_version(self, args):
        for spec in args:
            name, eq, version = spec.partition('=')
            try:
                if eq != '':
                    yield self.get_version(name, version.lstrip('='))
                elif self.all_versions:
                    p = self.get_package(name)
                    for v in sorted(p["releases"], key=parse):
                        if self.pre or not parse(v).is_prerelease:
                            if v == p["info"]["version"]:
                                yield p
                            else:
                                ### TODO: Can this call ever fail?
                                yield self.get_version(name, v)
                else:
                    yield self.get_latest_version(name)
            except QyPIError as e:
                self.errmsgs.append(str(e))

    def cleanup(self, ctx):
        if self.errmsgs:
            for msg in self.errmsgs:
                click.echo(ctx.command_path + ': ' + msg, err=True)
            ctx.exit(1)


class QyPIError(Exception):
    pass


def first_upload(files):
    return min((f["upload_time"] for f in files), default=None)



"""
Query PyPI from the command line

``qypi`` is a command-line client for querying & searching `PyPI
<https://pypi.org>`_ for Python package information and outputting JSON (with
some minor opinionated changes to the output data structures).

Run ``qypi --help`` or visit <https://github.com/jwodder/qypi> for more
information.
"""

__version__      = '0.5.0'
__author__       = 'John Thorvald Wodder II'
__author_email__ = 'qypi@varonathe.org'
__license__      = 'MIT'
__url__          = 'https://github.com/jwodder/qypi'



import click
from   packaging.version import parse
from   .                 import __version__
from   .api              import QyPI, first_upload
from   .util             import JSONLister, JSONMapper, clean_pypi_dict, \
                                    dumps, package_args, squish_versions

#ENDPOINT = 'https://pypi.python.org/pypi'
ENDPOINT = 'https://pypi.org/pypi'
TRUST_DOWNLOADS = False

SEARCH_SYNONYMS = {
    'homepage': 'home_page',
    'url': 'home_page',
    'long_description': 'description',
    'readme': 'description',
    'keyword': 'keywords',
}

@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.option('-i', '--index-url', default=ENDPOINT, metavar='URL',
              help='Use a different URL for PyPI', show_default=True)
@click.version_option(__version__, '-V', '--version',
                      message='%(prog)s %(version)s')
@click.pass_context
def qypi(ctx, index_url):
    """ Query PyPI from the command line """
    ctx.obj = QyPI(index_url)

@qypi.resultcallback()
@click.pass_context
def cleanup(ctx, *args, **kwargs):
    ctx.obj.cleanup(ctx)

@qypi.command()
@click.option('--description/--no-description', default=False,
              help='Include (long) descriptions', show_default=True)
@click.option('--trust-downloads/--no-trust-downloads', default=TRUST_DOWNLOADS,
              help='Show download stats', show_default=True)
@package_args()
def info(packages, trust_downloads, description):
    """
    Show package details.

    Packages can be specified as either ``packagename`` to show the latest
    version or as ``packagename==version`` to show the details for ``version``.
    """
    with JSONLister() as jlist:
        for pkg in packages:
            info = clean_pypi_dict(pkg["info"])
            if not description:
                info.pop('description', None)
            if not trust_downloads:
                info.pop('downloads', None)
            info["url"] = info.pop('home_page', None)
            info["release_date"] = first_upload(pkg["urls"])
            info["people"] = []
            for role in ('author', 'maintainer'):
                name = info.pop(role, None)
                email = info.pop(role + '_email', None)
                if name or email:
                    info["people"].append({
                        "name": name,
                        "email": email,
                        "role": role,
                    })
            if "package_url" in info and "project_url" not in info:
                # Field was renamed between PyPI Legacy and Warehouse
                info["project_url"] = info.pop("package_url")
            jlist.append(info)

@qypi.command()
@package_args()
def readme(packages):
    """
    View packages' long descriptions.

    If stdout is a terminal, the descriptions are passed to a pager program
    (e.g., `less(1)`).

    Packages can be specified as either ``packagename`` to show the latest
    version or as ``packagename==version`` to show the long description for
    ``version``.
    """
    for pkg in packages:
        click.echo_via_pager(pkg["info"]["description"])

@qypi.command()
@package_args(versioned=False)
def releases(packages):
    """ List released package versions """
    with JSONMapper() as jmap:
        for pkg in packages:
            try:
                project_url = pkg["info"]["project_url"]
            except KeyError:
                project_url = pkg["info"]["package_url"]
            if not project_url.endswith('/'):
                project_url += '/'
            jmap.append(
                pkg["info"]["name"],
                [{
                    "version": version,
                    "is_prerelease": parse(version).is_prerelease,
                    "release_date": first_upload(pkg["releases"][version]),
                    "release_url": project_url + version,
                } for version in sorted(pkg["releases"], key=parse)],
            )

@qypi.command()
@click.option('--trust-downloads/--no-trust-downloads', default=TRUST_DOWNLOADS,
              help='Show download stats', show_default=True)
@package_args()
def files(packages, trust_downloads):
    """
    List files available for download.

    Packages can be specified as either ``packagename`` to show the latest
    version or as ``packagename==version`` to show the files available for
    ``version``.
    """
    with JSONLister() as jlist:
        for pkg in packages:
            pkgfiles = pkg["urls"]
            for pf in pkgfiles:
                if not trust_downloads:
                    pf.pop("downloads", None)
                pf.pop("path", None)
                ### Change empty comment_text fields to None?
            jlist.append({
                "name": pkg["info"]["name"],
                "version": pkg["info"]["version"],
                "files": pkgfiles,
            })

@qypi.command('list')
@click.pass_obj
def listcmd(obj):
    """ List all packages on PyPI """
    for pkg in obj.xmlrpc('list_packages'):
        click.echo(pkg)

@qypi.command()
@click.option('--and', 'oper', flag_value='and', default=True,
              help='AND conditions together [default]')
@click.option('--or', 'oper', flag_value='or', help='OR conditions together')
@click.option('-p/-r', '--packages/--releases', default=False,
              help='Show one result per package/per release'
                   ' [default: per release]')
@click.argument('terms', nargs=-1, required=True)
@click.pass_obj
def search(obj, terms, oper, packages):
    """
    Search PyPI for packages or releases thereof.

    Search terms may be specified as either ``field:value`` (e.g.,
    ``summary:Django``) or just ``value`` to search long descriptions.
    """
    spec = {}
    for t in terms:
        key, colon, value = t.partition(':')
        if colon == '':
            key, value = 'description', t
        else:
            key = SEARCH_SYNONYMS.get(key, key)
        # ServerProxy can't handle defaultdicts, so we can't use those instead.
        spec.setdefault(key, []).append(value)
    results = map(clean_pypi_dict, obj.xmlrpc('search', spec, oper))
    if packages:
        results = squish_versions(results)
    click.echo(dumps(results))

@qypi.command()
@click.option('-f', '--file', type=click.File('r'))
@click.option('-p/-r', '--packages/--releases', default=False,
              help='Show one result per package/per release'
                   ' [default: per release]')
@click.argument('classifiers', nargs=-1)
@click.pass_obj
def browse(obj, classifiers, file, packages):
    """
    List packages with given trove classifiers.

    The list of classifiers may optionally be read from a file, one classifier
    per line.  Any further classifiers listed on the command line will be added
    to the file's list.
    """
    if file is not None:
        classifiers += tuple(map(str.strip, file))
    results = [
        {"name": name, "version": version or None}
        for name, version in obj.xmlrpc('browse', classifiers)
    ]
    if packages:
        results = squish_versions(results)
    click.echo(dumps(results))

@qypi.command()
@package_args(versioned=False)
# Map through the JSON API so we can get the correct casing to query the
# XML-RPC API with
@click.pass_obj
def owner(obj, packages):
    """ List package owners & maintainers """
    with JSONMapper() as jmap:
        for pkg in packages:
            name = pkg["info"]["name"]
            jmap.append(name, [
                {"role": role, "user": user}
                for role, user in obj.xmlrpc('package_roles', name)
            ])

@qypi.command()
@click.argument('users', nargs=-1)
@click.pass_obj
def owned(obj, users):
    """ List packages owned/maintained by a user """
    with JSONMapper() as jmap:
        for u in users:
            jmap.append(u, [
                {"role": role, "package": pkg}
                for role, pkg in obj.xmlrpc('user_packages', u)
            ])

if __name__ == '__main__':
    qypi()
